@@ -32,7 +32,7 @@ extern "C" {
 #include <spa/utils/defs.h>
 #include <spa/utils/list.h>
 
-/** \defgroup spa_interfaces SPA Interfaces
+/** \defgroup spa_interfaces Interfaces
  *
  * \brief Generic implementation of implementation-independent interfaces
  *
@@ -171,12 +171,15 @@ struct spa_interface {
 /**
  * Invoke method named \a method in the \a callbacks.
  * The \a method_type defines the type of the method struct.
+ * Returns true if the method could be called, false otherwise.
  */
 #define spa_callbacks_call(callbacks,type,method,vers,...)			\
 ({										\
 	const type *_f = (const type *) (callbacks)->funcs;			\
-	if (SPA_LIKELY(SPA_CALLBACK_CHECK(_f,method,vers)))			\
+	bool _res = SPA_CALLBACK_CHECK(_f,method,vers);				\
+	if (SPA_LIKELY(_res))							\
 		_f->method((callbacks)->data, ## __VA_ARGS__);			\
+	_res;									\
 })
 
 /**
@@ -188,6 +191,16 @@ struct spa_interface {
 	SPA_CALLBACK_VERSION_MIN(_f,vers);					\
 })
 
+/**
+ * True if the \a callbacks contains \a method of version
+ * \a vers, false otherwise
+ */
+#define spa_callback_check(callbacks,type,method,vers)				\
+({										\
+	const type *_f = (const type *) (callbacks)->funcs;			\
+	SPA_CALLBACK_CHECK(_f,method,vers);					\
+})
+
 /**
  * Invoke method named \a method in the \a callbacks.
  * The \a method_type defines the type of the method struct.
@@ -203,17 +216,24 @@ struct spa_interface {
 })
 
 /**
- * True if the \a iface's \a callbacks are of version \a vers, false otherwise
+ * True if the \a iface's callbacks are of version \a vers, false otherwise
  */
-#define spa_interface_callback_version_min(iface,method_type,vers)				\
+#define spa_interface_callback_version_min(iface,method_type,vers)		\
    spa_callback_version_min(&(iface)->cb, method_type, vers)
 
+/**
+ * True if the \a iface's callback \a method is of version \a vers
+ * and exists, false otherwise
+ */
+#define spa_interface_callback_check(iface,method_type,method,vers)		\
+   spa_callback_check(&(iface)->cb, method_type, method, vers)
+
 /**
  * Invoke method named \a method in the callbacks on the given interface object.
  * The \a method_type defines the type of the method struct, not the interface
  * itself.
  */
-#define spa_interface_call(iface,method_type,method,vers,...)				\
+#define spa_interface_call(iface,method_type,method,vers,...)			\
 	spa_callbacks_call(&(iface)->cb,method_type,method,vers,##__VA_ARGS__)
 
 /**
@@ -230,7 +250,7 @@ struct spa_interface {
  * \}
  */
 
-/** \defgroup spa_hooks SPA Hooks
+/** \defgroup spa_hooks Hooks
  *
  * A SPA Hook is a data structure to keep track of callbacks. It is similar to
  * the \ref spa_interfaces and typically used where an implementation allows
@@ -241,8 +261,10 @@ struct spa_interface {
  * The below (pseudo)code is a minimal example outlining the use of hooks:
  * \code{.c}
  * // the public interface
+ * #define VERSION_BAR_EVENTS 0 // version of the vtable
  * struct bar_events {
- *    uint32_t version;
+ *    uint32_t version; // NOTE: an integral member named `version`
+ *                      //       must be present in the vtable
  *    void (*boom)(void *data, const char *msg);
  * };
  *
@@ -252,17 +274,23 @@ struct spa_interface {
  * };
  *
  * void party_add_event_listener(struct party *p, struct spa_hook *listener,
- *                               struct bar_events *events, void *data)
+ *                               const struct bar_events *events, void *data)
  * {
  *    spa_hook_list_append(&p->bar_list, listener, events, data);
  * }
  *
  * static void party_on(struct party *p)
  * {
- *     spa_hook_list_call(&p->list, struct bar_events,
- *                        boom, // function name
- *                        0 // hardcoded version,
- *                        "party on, wayne");
+ *     // NOTE: this is a macro, it evaluates to an integer,
+ *     //       which is the number of hooks called
+ *     spa_hook_list_call(&p->list,
+ *                        struct bar_events, // vtable type
+ *                        boom,              // function name
+ *                        0,                 // hardcoded version,
+ *                                           //     usually the version in which `boom`
+ *                                           //     has been added to the vtable
+ *                        "party on, wayne"  // function argument(s)
+ *                        );
  * }
  * \endcode
  *
@@ -273,7 +301,8 @@ struct spa_interface {
  *      printf("%s", msg);
  * }
  *
- * static const struct bar_events {
+ * static const struct bar_events events = {
+ *    .version = VERSION_BAR_EVENTS, // version of the implemented interface
  *    .boom = boom_cb,
  * };
  *
@@ -282,7 +311,7 @@ struct spa_interface {
  *      struct spa_hook hook;
  *      struct party *p = start_the_party();
  *
- *      party_add_event_listener(p, &hook, boom_cb, userdata);
+ *      party_add_event_listener(p, &hook, &events, userdata);
  *
  *      mainloop();
  *      return 0;
@@ -400,22 +429,20 @@ spa_hook_list_join(struct spa_hook_list *list,
  * called */
 #define spa_hook_list_do_call(l,start,type,method,vers,once,...)		\
 ({										\
-	struct spa_hook_list *list = l;						\
-	struct spa_list *s = start ? (struct spa_list *)start : &list->list;	\
-	struct spa_hook cursor = { 0 }, *ci;					\
-	int count = 0;								\
-	spa_list_cursor_start(cursor, s, link);					\
-	spa_list_for_each_cursor(ci, cursor, &list->list, link) {		\
-		const type *_f = (const type *)ci->cb.funcs;			\
-		if (SPA_LIKELY(SPA_CALLBACK_CHECK(_f,method,vers))) {		\
-			_f->method(ci->cb.data, ## __VA_ARGS__);		\
-			count++;						\
+	struct spa_hook_list *_list = l;					\
+	struct spa_list *_s = start ? (struct spa_list *)start : &_list->list;	\
+	struct spa_hook _cursor = { 0 }, *_ci;					\
+	int _count = 0;								\
+	spa_list_cursor_start(_cursor, _s, link);				\
+	spa_list_for_each_cursor(_ci, _cursor, &_list->list, link) {		\
+		if (spa_callbacks_call(&_ci->cb,type,method,vers, ## __VA_ARGS__)) {		\
+			_count++;						\
 			if (once)						\
 				break;						\
 		}								\
 	}									\
-	spa_list_cursor_end(cursor, link);					\
-	count;									\
+	spa_list_cursor_end(_cursor, link);					\
+	_count;									\
 })
 
 /**
